/***************************************************************************
 *   Copyright (C) 2005 by Robot Group Leipzig                             *
 *    martius@informatik.uni-leipzig.de                                    *
 *    fhesse@informatik.uni-leipzig.de                                     *
 *    der@informatik.uni-leipzig.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *
 ***************************************************************************/

#include <selforg/noisegenerator.h>

#include <ode_robots/simulation.h>
#include <ode_robots/odeagent.h>
#include <selforg/one2onewiring.h>
#include <selforg/motornoisewiring.h>
#include <ode_robots/playground.h>
#include <ode_robots/passivebox.h>

#include <selforg/onecontrollerperchannel.h>
#include <selforg/sinecontroller.h>
#include <selforg/sos.h>
#include "pimax.h"

#include <ode_robots/sliderwheelie.h>

using namespace lpzrobots;

bool   track        = false;
bool   tracksegm    = false;
bool   useNoiseCtrl = false;
double noisecorrlen = 1;
double motornoise   = 1.0;
double eps          = 0.1;
double noise        = 0;
double gravity      = -6;
double initC        = 0.01;
bool   env          = false;

struct ControlGen : public ControllerGenerator {

  ControlGen(double eps):eps(eps){}
  virtual ~ControlGen(){}
  virtual AbstractController* operator()( int index) {
    AbstractController* c;
    c= new Sos(0.01);
    c->setParam("epsC",eps);
    c->setParam("epsA",.1);
    return c;
 }
  double eps;
};

struct ControlGenPiMax : public ControllerGenerator {
  ControlGenPiMax(double eps) :eps(eps) {}
  virtual ~ControlGenPiMax() {}
  virtual AbstractController* operator()( int index) {
    AbstractController* c;
    PiMaxConf pc = PiMax::getDefaultConf();
    pc.initFeedbackStrength=initC;
    pc.useExtendedModel=false;
    pc.onlyMainParameters=false;
    c= new PiMax(pc);
    c->setParam("epsC",eps);
    c->setParam("epsA",.005);
    c->setParam("sense",20);
    c->setParam("epsSigma",0.001);
    return c;
  }
  double eps;
};



class ThisSim : public Simulation {
public:

  AbstractController* controller;

  /// start() is called at the start and should create all the object (obstacles, agents...).
  virtual void start(const OdeHandle& odeHandle, const OsgHandle& osgHandle, GlobalData& global){

    setCameraHomePos(Pos(1.86452, 21.1998, 8.7964),  Pos(174.993, -10.688, 0));
    setCameraMode(Follow);

    global.odeConfig.setParam("controlinterval",4);
    global.odeConfig.setParam("gravity", gravity);

    global.odeConfig.setParam("noise", noise);
    //    global.odeConfig.setParam("realtimefactor", 1.);

    int sliderwheelies = 1;

    if(env){
      Playground* playground = new Playground(odeHandle, osgHandle,
                                              osg::Vec3(50, 0.2, 1.0),1,true);
      //    playground->setColor(Color(1,0.2,0,0.1));
      playground->setPosition(osg::Vec3(0,0,0)); // playground positionieren und generieren
      global.obstacles.push_back(playground);

      // Creation of Obstacle
      PassiveBox* b =
        new  PassiveBox(odeHandle,
                        osgHandle.changeColor(Color(180.0 / 255.0, 180.0 / 255.0, 130.0 / 255.0)),
                        osg::Vec3(2.0, 10.0, 0.4), 0);
      b->setTexture(0,TextureDescr("Images/playfulmachines.rgb",1,1));
      b->setPosition(Pos(-18, 0, 0));
      global.obstacles.push_back(b);
    }


    /******* S L I D E R - W H E E L I E *********/
    for(int i=0; i<sliderwheelies; i++){
      SliderWheelieConf mySliderWheelieConf = SliderWheelie::getDefaultConf();
      mySliderWheelieConf.segmNumber         = 12;
      mySliderWheelieConf.segmLength         = .4;
      mySliderWheelieConf.segmDia            = .1; // thickness and width(*8) of segments
      mySliderWheelieConf.segmLength         = .6;
      mySliderWheelieConf.segmDia            = .2; // thickness and width(*8) of segments
      mySliderWheelieConf.jointLimitIn       = M_PI/3;
      mySliderWheelieConf.frictionGround     = 0.5;
      mySliderWheelieConf.motorPower         = 5;
      mySliderWheelieConf.motorDamp          = 0.05;
      mySliderWheelieConf.sliderLength       = 0.5;

      OdeRobot* robot = new SliderWheelie(odeHandle,
                                          osgHandle.changeColor("Weissgrau"),
                                          mySliderWheelieConf, "SliderArmband_e"
                                          + std::itos(eps*10000,"%05i"));
      robot->place(Pos(0,0,2.0));


      OdeAgent* agent = new OdeAgent(global);

      AbstractWiring* wiring;
      if(useNoiseCtrl){
        assert(noisecorrlen>0);
        controller = new SineController();
        controller->setParam("amplitude",0);
        wiring = new MotorNoiseWiring(new ColorUniformNoise(1/(noisecorrlen)),motornoise);
      }else{
        // using homeokinesis controller
        //controller = new OneControllerPerChannel(new ControlGen(2.0),"OnePerJoint");
        // using pimax controller
        controller = new OneControllerPerChannel(new ControlGenPiMax(eps),"OnePerJoint",
                                                 mySliderWheelieConf.segmNumber*3/2);
        wiring = new One2OneWiring(new WhiteUniformNoise());
        //        wiring = new One2OneWiring(new ColorUniformNoise(0.1));
        // only the first controller is exported to guilogger & Co
        // agent->addInspectable(((OneControllerPerChannel*)controller)->getControllers()[0]);
        // all controller are exported to guilogger & Co
        for(int i=0; i< mySliderWheelieConf.segmNumber*3/2; i++){
          agent->addInspectable(((OneControllerPerChannel*)controller)->getControllers()[i]);
        }
      }

      agent->init(controller, robot, wiring);
      if(track){
        TrackRobotConf c = TrackRobot::getDefaultConf();
        c.displayTrace = true;
        c.scene        = "pi_split_control";
        c.interval     = 1;
        c.trackSpeed   = true;
        agent->setTrackOptions(TrackRobot(c));
        agent->addConfigurable(&global.odeConfig);
      }
      if(tracksegm){
        TrackRobotConf c = TrackRobot::getDefaultConf();
        Color          col(1,0,0,0.5);
        c.displayTrace = true;
        c.displayTraceThickness = 0.02;
        c.displayTraceDur = 10;
        c.scene        = "segm";
        c.interval     = 1;
        c.writeFile    = true;
        for(int i=0; i<mySliderWheelieConf.segmNumber/2; i++){
          agent->addTracking(i*3+2,TrackRobot(c), col);
        }
      }
      global.agents.push_back(agent);
      global.configs.push_back(agent);
    }
  }

  virtual void usage() const {
    printf("\t-motornoise strength\tcontrol robot by noise with given strength\n");
    printf("\t-corrlen tau \tuse noise as motor commands with correlation length 1/tau \n");
    printf("\t-noise strength\tnoise strength for sensor noise (def: 0.0)\n");
    printf("\t-eps learningrate\t learning rate for controller (def: 0.1)\n");
    printf("\t-gravity gravity\tvalue of gravity (def: -0.6)\n");
    printf("\t-cInit strength\tinitial value for Controller diagonal (def: 1.0)\n");
    printf("\t-track\ttrack center of robot\n");
    printf("\t-tracksegm\ttrack segments of robot\n");
    printf("\t-env\t put the robot in an environment\n");
  }
};

int main (int argc, char **argv)
{
  ThisSim sim;
  int index;

  index = sim.contains(argv,argc,"-motornoise");
  if(index >0 && index < argc){
    useNoiseCtrl=true;
    motornoise= atof(argv[index]);
  }
  index = sim.contains(argv,argc,"-corrlen");
  if(index >0 && index < argc){
    useNoiseCtrl=true;
    noisecorrlen= atof(argv[index]);
  }
  index = sim.contains(argv,argc,"-eps");
  if(index >0 && index < argc){
    eps= atof(argv[index]);
  }
  index = sim.contains(argv,argc,"-gravity");
  if(index >0 && index < argc){
    gravity= atof(argv[index]);
  }
  index = sim.contains(argv,argc,"-initC");
  if(index >0 && index < argc){
    initC= atof(argv[index]);
  }
  index = sim.contains(argv,argc,"-noise");
  if(index >0 && index < argc){
    noise= atof(argv[index]);
  }

  track     = sim.contains(argv,argc,"-track") != 0;
  tracksegm = sim.contains(argv,argc,"-tracksegm") != 0;
  env       = sim.contains(argv,argc,"-env") != 0;
  sim.setTitle("Pred.inf.maximization - split ctrl.       ");
  return sim.run(argc, argv) ? 0 : 1;
}

// for the plots
//for R in 0 1 2 3 4 5 6 7 8 9; do
//  for E in 0 0.0001 0.0002 0.0003 0.0005 0.001 0.002 0.005 0.01 0.02 0.03 0.05 0.07 0.1 0.2 0.3 0.5 1.0 2.0; do
//    nice ./start_opt -eps $E -noise 0.001 -initC 0.5 -track -simtime 10 -nographics;
// done;
//done

// for the video
// ./start -eps 0.5  -env -noise 0.0 -initC 0.5;
